@@ -336,7 +336,7 @@ static int calc_root_size(struct btrfs_root *tree_root, struct btrfs_key *key,
 	stat.max_cluster_size = root->fs_info->nodesize;
 	path.nodes[level] = root->node;
 	if (gettimeofday(&start, NULL)) {
-		error("cannot get time: %s", strerror(errno));
+		error("cannot get time: %m");
 		goto out;
 	}
 	if (!level) {
@@ -350,7 +350,7 @@ static int calc_root_size(struct btrfs_root *tree_root, struct btrfs_key *key,
 	if (ret)
 		goto out;
 	if (gettimeofday(&end, NULL)) {
-		error("cannot get time: %s", strerror(errno));
+		error("cannot get time: %m");
 		goto out;
 	}
 	timeval_subtract(&diff, &end, &start);
@@ -434,6 +434,7 @@ int cmd_inspect_tree_stats(int argc, char **argv)
 	int opt;
 	int ret = 0;
 
+	optind = 0;
 	while ((opt = getopt(argc, argv, "vb")) != -1) {
 		switch (opt) {
 		case 'v':
@@ -443,18 +444,17 @@ int cmd_inspect_tree_stats(int argc, char **argv)
 			no_pretty = 1;
 			break;
 		default:
-			usage(cmd_inspect_tree_stats_usage);
+			usage_unknown_option(cmd_inspect_tree_stats_usage, argv);
 		}
 	}
 
-	if (check_argc_exact(argc - optind, 1)) {
-		usage(cmd_inspect_tree_stats_usage);
-	}
+	if (check_argc_exact(argc - optind, 1))
+		return 1;
 
 	ret = check_mounted(argv[optind]);
 	if (ret < 0) {
-		warning("unable to check mount status of: %s",
-				strerror(-ret));
+		errno = -ret;
+		warning("unable to check mount status of: %m");
 	} else if (ret) {
 		warning("%s already mounted, results may be inaccurate",
 				argv[optind]);
