@@ -42,6 +42,7 @@
 #include "rbtree-utils.h"
 #include "mkfs/common.h"
 #include "mkfs/rootdir.h"
+#include "kernel-lib/crc32c.h"
 #include "fsfeatures.h"
 
 static int verbose = 1;
@@ -811,6 +812,8 @@ int main(int argc, char **argv)
 	struct mkfs_allocation allocation = { 0 };
 	struct btrfs_mkfs_config mkfs_cfg;
 
+	crc32c_optimization_init();
+
 	while(1) {
 		int c;
 		enum { GETOPT_VAL_SHRINK = 257 };
